@@ -4,14 +4,18 @@ from pathlib import Path
 import sys
 
 import click
+import humanize
 from tabulate import tabulate
 
-from . import AliasGroup, main
-from .pretty import print_wait, print_done, print_error, print_fail
+from . import main
+from .interaction import ask_yn
+from .pretty import print_done, print_error, print_fail, print_info, print_wait
+from .utils import ByteSizeParamType, ByteSizeParamCheckType
+from ..config import DEFAULT_CHUNK_SIZE
 from ..session import Session
 
 
-@main.group(cls=AliasGroup)
+@main.group()
 def vfolder():
     '''Provides virtual folder operations.'''
 
@@ -24,11 +28,14 @@ def list(list_all):
     fields = [
         ('Name', 'name'),
         ('ID', 'id'),
+        ('Volume', 'host'),
         ('Owner', 'is_owner'),
         ('Permission', 'permission'),
-        ('Type', 'type'),
+        ('Owership Type', 'ownership_type'),
+        ('Usage Mode', 'usage_mode'),
         ('User', 'user'),
         ('Group', 'group'),
+        ('Created At', 'created_at'),
     ]
     with Session() as session:
         try:
@@ -74,7 +81,24 @@ def list_allowed_types():
 @click.argument('host', type=str, default=None)
 @click.option('-g', '--group', metavar='GROUP', type=str, default=None,
               help='Group ID or NAME. Specify this option if you want to create a group folder.')
-def create(name, host, group):
+@click.option('--unmanaged', 'host_path', type=bool, is_flag=True,
+              help='Treats HOST as a mount point of unmanaged virtual folder. '
+                   'This option can only be used by Admin or Superadmin.')
+@click.option('-m', '--usage-mode', metavar='USAGE_MODE', type=str, default='general',
+              help='Purpose of the folder. Normal folders are usually set to "general". '
+                   'Available options: "general", "data" (provides data to users), '
+                   'and "model" (provides pre-trained models).')
+@click.option('-p', '--permission', metavar='PERMISSION', type=str, default='rw',
+              help='Folder\'s innate permission. '
+                   'Group folders can be shared as read-only by setting this option to "ro".'
+                   'Invited folders override this setting by its own invitation permission.')
+@click.option('-q', '--quota', metavar='QUOTA', type=ByteSizeParamCheckType(), default='0',
+              help='Quota of the virtual folder. '
+                   '(Use \'m\' for megabytes, \'g\' for gigabytes, and etc.) '
+                   'Default is maximum amount possible.')
+@click.option('--cloneable', '--allow-clone', type=bool, is_flag=True,
+              help='Allows the virtual folder to be cloned by users.')
+def create(name, host, group, host_path, usage_mode, permission, quota, cloneable):
     '''Create a new virtual folder.
 
     \b
@@ -83,7 +107,26 @@ def create(name, host, group):
     '''
     with Session() as session:
         try:
-            result = session.VFolder.create(name, host, group)
+            if host_path:
+                result = session.VFolder.create(
+                    name=name,
+                    unmanaged_path=host,
+                    group=group,
+                    usage_mode=usage_mode,
+                    permission=permission,
+                    quota=quota,
+                    cloneable=cloneable,
+                )
+            else:
+                result = session.VFolder.create(
+                    name=name,
+                    host=host,
+                    group=group,
+                    usage_mode=usage_mode,
+                    permission=permission,
+                    quota=quota,
+                    cloneable=cloneable,
+                )
             print('Virtual folder "{0}" is created.'.format(result['name']))
         except Exception as e:
             print_error(e)
@@ -142,20 +185,31 @@ def info(name):
             print('- Owner:', result['is_owner'])
             print('- Permission:', result['permission'])
             print('- Number of files: {0}'.format(result['numFiles']))
-            print('- Type: {0}'.format(result['type']))
+            print('- Ownership Type: {0}'.format(result['type']))
+            print('- Permission:', result['permission'])
+            print('- Usage Mode: {0}'.format(result.get('usage_mode', '')))
             print('- Group ID: {0}'.format(result['group']))
             print('- User ID: {0}'.format(result['user']))
+            print('- Clone Allowed: {0}'.format(result['cloneable']))
         except Exception as e:
             print_error(e)
             sys.exit(1)
 
 
-@vfolder.command()
+@vfolder.command(context_settings={'show_default': True})  # bug: pallets/click#1565 (fixed in 8.0)
 @click.argument('name', type=str)
 @click.argument('filenames', type=Path, nargs=-1)
-def upload(name, filenames):
+@click.option('-b', '--base-dir', type=Path, default=None,
+              help='Set the parent directory from where the file is uploaded.  '
+                   '[default: current working directry]')
+@click.option('--chunk-size', type=ByteSizeParamType(),
+              default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
+              help='Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
+                   'Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) '
+                   'and networks (e.g., 40 GbE) for the maximum throughput.')
+def upload(name, filenames, base_dir, chunk_size):
     '''
-    Upload a file to the virtual folder from the current working directory.
+    TUS Upload a file to the virtual folder from the current working directory.
     The files with the same names will be overwirtten.
 
     \b
@@ -164,17 +218,30 @@ def upload(name, filenames):
     '''
     with Session() as session:
         try:
-            session.VFolder(name).upload(filenames, show_progress=True)
+            session.VFolder(name).upload(
+                filenames,
+                basedir=base_dir,
+                chunk_size=chunk_size,
+                show_progress=True,
+            )
             print_done('Done.')
         except Exception as e:
             print_error(e)
             sys.exit(1)
 
 
-@vfolder.command()
+@vfolder.command(context_settings={'show_default': True})  # bug: pallets/click#1565 (fixed in 8.0)
 @click.argument('name', type=str)
 @click.argument('filenames', type=Path, nargs=-1)
-def download(name, filenames):
+@click.option('-b', '--base-dir', type=Path, default=None,
+              help='Set the parent directory from where the file is uploaded.  '
+                   '[default: current working directry]')
+@click.option('--chunk-size', type=ByteSizeParamType(),
+              default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
+              help='Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
+                   'Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) '
+                   'and networks (e.g., 40 GbE) for the maximum throughput.')
+def download(name, filenames, base_dir, chunk_size):
     '''
     Download a file from the virtual folder to the current working directory.
     The files with the same names will be overwirtten.
@@ -185,13 +252,38 @@ def download(name, filenames):
     '''
     with Session() as session:
         try:
-            session.VFolder(name).download(filenames, show_progress=True)
+            session.VFolder(name).download(
+                filenames,
+                basedir=base_dir,
+                chunk_size=chunk_size,
+                show_progress=True,
+            )
             print_done('Done.')
         except Exception as e:
             print_error(e)
             sys.exit(1)
 
 
+@vfolder.command()
+@click.argument('name', type=str)
+@click.argument('filename', type=Path)
+def request_download(name, filename):
+    '''
+    Request JWT-formated download token for later use.
+
+    \b
+    NAME: Name of a virtual folder.
+    FILENAME: Path of the file to be downloaded.
+    '''
+    with Session() as session:
+        try:
+            response = json.loads(session.VFolder(name).request_download(filename))
+            print_done(f'Download token: {response["token"]}')
+        except Exception as e:
+            print_error(e)
+            sys.exit(1)
+
+
 @vfolder.command()
 @click.argument('filenames', nargs=-1)
 def cp(filenames):
@@ -266,11 +358,13 @@ def rm(name, filenames, recursive):
     '''
     with Session() as session:
         try:
-            if input("> Are you sure? (y/n): ").lower().strip()[:1] == 'y':
-                session.VFolder(name).delete_files(
-                    filenames,
-                    recursive=recursive)
-                print_done('Done.')
+            if not ask_yn():
+                print_info('Cancelled')
+                sys.exit(1)
+            session.VFolder(name).delete_files(
+                filenames,
+                recursive=recursive)
+            print_done('Done.')
         except Exception as e:
             print_error(e)
             sys.exit(1)
@@ -392,3 +486,94 @@ def invitations():
         except Exception as e:
             print_error(e)
             sys.exit(1)
+
+
+@vfolder.command()
+@click.argument('name', type=str)
+def leave(name):
+    '''Leave the shared virutal folder.
+
+    NAME: Name of a virtual folder
+    '''
+    with Session() as session:
+        try:
+            vfolder_info = session.VFolder(name).info()
+            if vfolder_info['type'] == 'group':
+                print('You cannot leave a group virtual folder.')
+                return
+            if vfolder_info['is_owner']:
+                print('You cannot leave a virtual folder you own. Consider using delete instead.')
+                return
+            session.VFolder(name).leave()
+            print('Left the shared virtual folder "{}".'.format(name))
+
+        except Exception as e:
+            print_error(e)
+            sys.exit(1)
+
+
+@vfolder.command()
+@click.argument('name', type=str)
+@click.argument('target_name', type=str)
+@click.argument('target_host', type=str)
+@click.option('-m', '--usage-mode', metavar='USAGE_MODE', type=str, default='general',
+              help='Purpose of the cloned virtual folder. '
+                   'Default value is \'general\'.')
+@click.option('-p', '--permission', metavar='PERMISSION', type=str, default='rw',
+              help='Cloned virtual folder\'s permission. '
+                   'Default value is \'rw\'.')
+def clone(name, target_name, target_host, usage_mode, permission):
+    """Clone a virtual folder.
+
+    \b
+    NAME: Name of the virtual folder to clone from.
+    TARGET_NAME: Name of the virtual folder to clone to.
+    TARGET_HOST: Name of a virtual folder host to which the virtual folder will be cloned.
+    """
+    with Session() as session:
+        try:
+            vfolder_info = session.VFolder(name).info()
+            if not vfolder_info['cloneable']:
+                print("Clone is not allowed for this virtual folder. "
+                      "Please update the 'cloneable' option.")
+                return
+            session.VFolder(name).clone(
+                target_name,
+                target_host=target_host,
+                usage_mode=usage_mode,
+                permission=permission,
+            )
+            print_done("Cloned.")
+        except Exception as e:
+            print_error(e)
+            sys.exit(1)
+
+
+@vfolder.command()
+@click.argument('name', type=str)
+@click.option('-p', '--permission', type=str, metavar='PERMISSION',
+              help="Folder's innate permission.")
+@click.option('--set-cloneable', type=bool, metavar='BOOLEXPR',
+              help="A boolean-interpretable string whether a virtual folder can be cloned. "
+                   "If not set, the cloneable property is not changed.")
+def update_options(name, permission, set_cloneable):
+    """Update an existing virtual folder.
+
+    \b
+    NAME: Name of the virtual folder to update.
+    """
+    with Session() as session:
+        try:
+            vfolder_info = session.VFolder(name).info()
+            if not vfolder_info['is_owner']:
+                print("You cannot update virtual folder that you do not own.")
+                return
+            session.VFolder(name).update_options(
+                name,
+                permission=permission,
+                cloneable=set_cloneable,
+            )
+            print_done("Updated.")
+        except Exception as e:
+            print_error(e)
+            sys.exit(1)
